"""Custom Bocadillo commands.

Use Click to build custom commands. For documentation, see:
https://click.palletsprojects.com
"""
import click


@click.group()
def cli():
    pass


# Write your @cli.command() functions below.
